import muuusiiik.util as msk
from   pytest import raises

# ============
# MOCK UP ENV
# ============
class MockEnv:
    env_path    = 'tests/_data'
    f_configure = f'{env_path}/system.conf'

    def create_configure(conf):
        msk.configure.save(conf, MockEnv.f_configure)

    def create_contents(contents):
        for k, v in contents.items():
            f = f'{MockEnv.env_path}/{k}'
            msk.data.save(v, f)

    def remove_env():
        msk.data.rm(MockEnv.env_path)


# =============
# CORE TESTING
# =============
def test_mock_content():
    # init env
    contents  = {
                  'file01.vocab': ['f1', 'f2', 'f3'],
                  'file02.vocab': ['v1', 'v2', 'v3', 'v4']
                }
    env_path  = MockEnv.env_path
    configure = {'vocab': [f'{env_path}/{k}' for k, v in contents.items()]}
    MockEnv.create_configure(configure)
    MockEnv.create_contents(contents)
    # assertation
    assert set(msk.data.ls(MockEnv.env_path)) == set(['file01.vocab', 'file02.vocab', 'system.conf'])
    # reset env
    MockEnv.remove_env()
    assert msk.data.exist(MockEnv.env_path) == False


def test_configure_loader_load_content():
    # init env
    contents  = {
                  'file01.vocab': ['f1', 'f2', 'f3'],
                  'file02.vocab': ['v1', 'v2', 'v3', 'v4']
                }
    env_path  = MockEnv.env_path
    configure = {'vocab': [f'{env_path}/{k}' for k, v in contents.items()]}
    MockEnv.create_configure(configure)
    MockEnv.create_contents(contents)
    # load configure
    env_path    = MockEnv.env_path
    f_configure = MockEnv.f_configure
    conf        = msk.configure_loader(f_configure)
    # assertation
    assert conf.path              == f_configure
    assert conf.content           == configure
    assert len(conf.get('vocab')) == 2
    assert [f.split('/')[-1] for f in conf.get('vocab')] == ['file01.vocab', 'file02.vocab']
    # reset env
    MockEnv.remove_env()


def test_configure_loader_load_non_existing_configure():
    f_configure = 'not-existing-file.conf'
    with raises(FileNotFoundError):
        conf = msk.configure_loader(f_configure)


def test_configure_loader_load_none_and_get_empty_class_as_reset():
    f_configure = None
    conf = msk.configure_loader(f_configure)
    assert conf.path    == None
    assert conf.content == {}




import muuusiiik.util as msk
from   pytest import raises


def test_there_is_a_log_file_when_set_filename():
    path   = 'tests/_data'
    f_log  = 'demo.log'
    f      = f'{path}/{f_log}'
    # make sure the path does not exist
    msk.data.rm(path)
    # make sure there is a folder
    assert msk.data.exist(path) == False
    msk.data.make_path(path, pathtype='folder')
    assert msk.data.exist(path) == True
    # make logger
    logger = msk.log.GetLogger(filename=f)
    logger.debug('hello')
    # check the log file is created
    assert msk.data.exist(f) == True
    # reset content
    msk.data.rm(path)
    assert msk.data.exist(f) == False


def test_there_is_a_log_file_when_set_filename_even_the_path_does_not_exist():
    path   = 'tests/_data'
    f_log  = 'demo.log'
    f      = f'{path}/{f_log}'
    # make sure the path does not exist
    msk.data.rm(path)
    assert msk.data.exist(path) == False
    # make logger
    logger = msk.log.GetLogger(filename=f)
    logger.debug('hello')
    # check the log file is created
    assert msk.data.exist(f) == True
    # reset content
    msk.data.rm(path)
    assert msk.data.exist(f) == False


def test_there_is_2_handlers_as_setting():
    path   = 'tests/_data'
    f_log  = 'tests/_data/demo.log'
    logger = msk.log.GetLogger()
    hand   = msk.log.GetHandler(filename=f_log)
    logger.addHandler(hand)
    assert len(logger.handlers) == 2
    # reset content
    msk.data.rm(path)
    assert msk.data.exist(f_log) == False


def test_logger_is_cleared_properly():
    name = 'demo_logger'
    # init root logger
    logger = msk.log.GetLogger(name)
    assert logger.name          == name
    assert len(logger.handlers) == 1
    # adding an extra handler
    hand   = msk.log.GetHandler()
    assert logger.name          == name
    logger.addHandler(hand)
    assert len(logger.handlers) == 2
    # clear logger
    msk.log.clear_by_name(name)
    assert logger.name          == name
    assert len(logger.handlers) == 0


def test_having_handler_only_one_even_init_it_twice():
    path   = 'tests/_data'
    f_log  = 'tests/_data/demo.log'
    # batch no.1
    logger = msk.log.GetLogger()
    hand   = msk.log.GetHandler(filename=f_log)
    logger.addHandler(hand)
    # batch no.2
    logger = msk.log.GetLogger()
    hand   = msk.log.GetHandler(filename=f_log)
    logger.addHandler(hand)
    # logger should not have 4 handlers
    assert len(logger.handlers) == 2
    # reset content
    msk.data.rm(path)
    assert msk.data.exist(f_log) == False


def test_formatter_validation():
    # general use
    formatter = 'nothing'
    assert msk.log._validate_formatter(formatter) == 'nothing'
    formatter = 'minimal'
    assert msk.log._validate_formatter(formatter) == 'minimal'

    # formatter is None
    formatter = None
    assert msk.log._validate_formatter(formatter) == 'minimal'
    # upper case
    formatter = 'NOTHING'
    assert msk.log._validate_formatter(formatter) == 'nothing'
    # not supporting value -> nothing
    formatter = 'non-exsisting'
    assert msk.log._validate_formatter(formatter) == 'nothing'


def test_when_validation():
    # daily and weekly
    when = 'daily'
    assert msk.log._validate_when(when) == 'midnight'
    when = 'weekly'
    assert msk.log._validate_when(when) == 'W0'

    # when is None
    when = None
    assert msk.log._validate_when(when) == when
    # when is any
    when = 'something-else'
    assert msk.log._validate_when(when) == when



import muuusiiik.util as msk
from   pytest import raises

def test_save_and_load_configure_by_dict():
    """ simply save and load a configure 
        the configure is in a dict format
    """
    conf = {'module': 'configure',
            'checklist': ['save', 'load']
           }
    path = 'tests/_config/demo.conf'
    msk.configure.save(conf, path)
    # the conf fire exist
    result = msk.data.exist(path)
    assert result == True
    # assertation
    loaded_conf = msk.configure.load(path)
    assert loaded_conf == conf
    # remove the demo folder
    fd, fn = msk.data.path_split(path)
    msk.data.rm(fd)
    result = msk.data.exist(path)
    assert result == False


def test_save_and_load_configure_by_plain_text():
    """ simply save and load a configure 
        the configure is in a yaml format
    """
    conf = """
    module: configure
    checklist:
    - save
    - load
    """
    path = 'tests/_config/demo.conf'
    msk.data.save(conf, path)
    # the conf fire exist
    result = msk.data.exist(path)
    assert result == True
    # assertation
    loaded_conf = msk.configure.load(path)
    assert loaded_conf['module'   ] == 'configure'
    assert loaded_conf['checklist'] == ['save', 'load']
    # remove the demo folder
    fd, fn = msk.data.path_split(path)
    msk.data.rm(fd)
    result = msk.data.exist(path)
    assert result == False


def test_load_non_existing_configure_file():
    """ scenario of oading a non-existing file
        expecting a file-not-found-error
    """
    path = 'tests/_config/non_existing.conf'
    with raises(FileNotFoundError):
        loaded_conf = msk.configure.load(path)


def test_load_none_configure_file():
    """ scenario of a given path is None
        expecting type-error
    """
    path = None
    with raises(TypeError):
        loaded_conf = msk.configure.load(path)



import muuusiiik.util as msk
import os

from   pytest import raises


# -----------------------------
# MOCK :: CREATE & REMOVE FILE
# -----------------------------
class Mock:
    def make_sure_there_is_a_file(file_name:str):
        fd, fn = msk.data.path_split(file_name)
        msk.data.make_path(fd)
        msk.data.save('', file_name)

    def make_sure_the_folder_removed(path:str):
        msk.data.rm(path)

    def make_a_new_folder_with_3_files_1_folder_inside(folder, fname):
        # mock a new folder with 3 files and 1 folder inside
        Mock.make_sure_the_folder_removed(folder)
        file_name = f'{folder}/{fname}'
        for i in range(3): Mock.make_sure_there_is_a_file(f'{file_name}_{i:02}.txt')
        msk.data.make_path(f'{folder}/not_exist_folder', pathtype='folder')

    def make_content_files(folder_path, contents):
        for k, v in contents.items():
            f = f'{folder_path}/{k}'
            msk.data.save(v, f)


# -----------------------------
# FILE PATH SPLIT 
# -----------------------------
def test_path_split_on_folder_path():
    file_name = '/some/folder_path/'
    fd, fn    = msk.data.path_split(file_name)
    assert (fd, fn) == ('/some/folder_path', '')
    

def test_path_split_on_file_path():
    file_name = '/some/folder_path/file_name.txt'
    fd, fn    = msk.data.path_split(file_name)
    assert (fd, fn) == ('/some/folder_path', 'file_name.txt')
    

def test_path_split_on_current_folder():
    # current path with filename
    file_name = 'file_name.txt'
    fd, fn    = msk.data.path_split(file_name)
    assert (fd, fn) == ('.', 'file_name.txt')

    # nothing in the path
    file_name = ''
    fd, fn    = msk.data.path_split(file_name)
    assert (fd, fn) == ('.', '')

    # current path is .
    file_name = '.'
    fd, fn    = msk.data.path_split(file_name)
    assert (fd, fn) == ('.', '')

    # file_name is single letter
    file_name = 'x'
    fd, fn    = msk.data.path_split(file_name)
    assert (fd, fn) == ('.', 'x')


def test_path_split_raise_attribute_error():
    with raises(AttributeError):
        filename = 1
        p, f     = msk.data.path_split(filename)


# -----------------------------
# CHECK TYPE OF FILE & FOLDER 
# -----------------------------
def test_path_type_is_file_when_give_file_path():
    path   = 'tests/test_data.py'
    result = msk.data.path_type(path)
    assert result == 'file'


def test_path_type_is_folder_when_give_folder_path():
    # with out slash
    path   = 'tests'
    result = msk.data.path_type(path)
    assert result == 'folder'

    # with slash
    path   = 'tests/'
    result = msk.data.path_type(path)
    assert result == 'folder'


def test_path_type_is_current_dir_represented_with_special_char():
    # path is dot
    path   = '.'
    result = msk.data.path_type(path)
    assert result == 'folder'

    # path is empty string
    path   = ''
    result = msk.data.path_type(path)
    assert result == 'folder'


def test_path_type_is_none_when_give_non_existing_path():
    path   = 'tests/non_existing_file.txt'
    result = msk.data.path_type(path)
    assert result == None


# -----------------------------
# CHECK IF FILE & FOLDER EXIST
# -----------------------------
def test_there_is_existing_file():
    file_name = 'tests/test_data.py'
    result    = msk.data.exist(file_name)
    assert result == True


def test_there_is_existing_file_without_folder_path():
    file_name = 'README.md'
    result    = msk.data.exist(file_name)
    assert result == True


def test_there_is_not_existing_file():
    file_name = 'tests/non_existing_file.py'
    result    = msk.data.exist(file_name)
    assert result == False


def test_there_is_existing_folder():
    # without slash
    folder_name = 'tests'
    result      = msk.data.exist(folder_name)
    assert result == True

    # with slash
    folder_name = 'tests/'
    result      = msk.data.exist(folder_name)
    assert result == True


def test_exist_current_folder_with_special_char():
    # current folder is dot
    folder_name = '.'
    result      = msk.data.exist(folder_name)
    assert result == True

    # current folder is empty string
    folder_name = ''
    result      = msk.data.exist(folder_name)
    assert result == True


def test_there_is_no_exising_folder():
    folder_name = 'not_existing_tests/'
    result      = msk.data.exist(folder_name)
    assert result == False


def test_raise_type_error_for_incorrect_path_type():
    with raises(TypeError):
        folder_name = 'not_existing_tests/'
        result      = msk.data.exist(folder_name, pathtype='document')


# -----------------------------
# LS A SPECIFIC PATH
# -----------------------------
def test_list_contents_in_a_given_folder_in_list_format():
    # mock a new folder with 3 files and 1 folder inside
    file_name = 'tests/_data/some_file'
    fd, fn    = msk.data.path_split(file_name)
    Mock.make_a_new_folder_with_3_files_1_folder_inside(fd, fn)

    # check ontents in given folder
    result    = msk.data.ls(fd)
    assert len(result) == 4
    assert ('not_exist_folder/' in result) == True
    assert ('some_file_00.txt'  in result) == True
    assert ('some_file_01.txt'  in result) == True
    assert ('some_file_02.txt'  in result) == True
    assert ('some_file_03.txt'  in result) == False

    # remove the folder
    Mock.make_sure_the_folder_removed(fd)
    assert msk.data.exist(fd) == False


def test_list_contents_in_a_given_folder_in_dict_format():
    # mock a new folder with 3 files and 1 folder inside
    file_name = 'tests/_data/some_file'
    fd, fn    = msk.data.path_split(file_name)
    Mock.make_a_new_folder_with_3_files_1_folder_inside(fd, fn)

    # check ontents in given folder
    result    = msk.data.ls(fd, fmt='dict')
    assert len(result) == 2
    assert len(result['folder']) == 1
    assert len(result['file'  ]) == 3
    assert ('not_exist_folder' in result['folder']) == True
    assert ('some_file_00.txt' in result['file'  ]) == True
    assert ('some_file_01.txt' in result['file'  ]) == True
    assert ('some_file_02.txt' in result['file'  ]) == True
    assert ('some_file_03.txt' in result['file'  ]) == False

    # remove the folder
    Mock.make_sure_the_folder_removed(fd)
    assert msk.data.exist(fd) == False


def test_list_contents_in_current_folder_with_special_char():
    path   = '.'
    result = msk.data.ls(path)
    assert ('README.md' in result)         == True
    assert ('not_existing_file' in result) == False

    path   = ''
    result = msk.data.ls(path)
    assert ('README.md' in result)         == True
    assert ('not_existing_file' in result) == False


def test_list_contents_in_a_non_existing_folder():
    path   = 'non_existing_path'
    with raises(FileNotFoundError):
        result = msk.data.ls(path)


def test_list_conents_in_a_path_the_same_as_file_name():
    path   = 'tests/test_data.py'
    with raises(NotADirectoryError):
        result = msk.data.ls(path)


# -----------------------------
# REMOVE FILE & FOLDER
# -----------------------------
def test_remove_a_file():
    # mock content 
    file_name = 'tests/_data/some_file'
    fd, fn    = msk.data.path_split(file_name)
    Mock.make_a_new_folder_with_3_files_1_folder_inside(fd, fn)

    # the mocked contents have 3 files and 1 folder
    ls_list   = msk.data.ls(fd)
    assert len(ls_list) == 4

    # remove file 01
    f         = f'{file_name}_01.txt'
    result    = msk.data.rm(f)
    ls_list   = msk.data.ls(fd)
    assert result         == True
    assert len(ls_list)   == 3
    assert (f in ls_list) == False

    # reset to zero
    Mock.make_sure_the_folder_removed(fd)
    assert msk.data.exist(fd) == False


def test_remove_a_folder():
    # mock content 
    file_name = 'tests/_data/some_file'
    fd, fn    = msk.data.path_split(file_name)
    Mock.make_a_new_folder_with_3_files_1_folder_inside(fd, fn)

    # remove tests/_data/
    result    = msk.data.rm(fd)
    assert result == True

    # folder tests/_data/ does not exist anymore in tests/
    ls_list   = msk.data.ls('tests')
    assert (fd in ls_list) == False


def test_remove_a_non_existing_file():
    file_name = 'tests/non_existing_file.txt'
    result    = msk.data.rm(file_name)
    assert result == False


def test_remove_a_non_existing_folder():
    folder_path = 'tests/non_existing_folder/'
    result      = msk.data.rm(folder_path)
    assert result == False


# -----------------------------
# MAKE PATH
# -----------------------------
def test_make_path_from_various_paths():
    # make sure tests/_data/ folder is not exist
    result = msk.data.rm('tests/_data')

    # #make a folder for testing this demo
    path   = 'tests/_data'
    result = msk.data.make_path(path, pathtype='folder')
    assert result == True

    # make a folder from file_path
    path   = 'tests/_data/non_existing_folder_01/with_file.txt'
    result = msk.data.make_path(path)
    assert result == True

    # make a folder from folder_path with ending slash, pathtype=file
    path   = 'tests/_data/non_existing_folder_02/'
    result = msk.data.make_path(path, pathtype='file')
    assert result == True

    # make a folder from folder_path with ending slash, pathtype=folder
    path   = 'tests/_data/non_existing_folder_03/'
    result = msk.data.make_path(path, pathtype='folder')
    assert result == True

    # make a folder from folder_path WITHOUT ending slash, pathtype=file >> do not make a folder
    path   = 'tests/_data/non_existing_folder_04'
    result = msk.data.make_path(path, pathtype='file')
    assert result == True

    # make a folder from folder_path WITHOUT ending slash, pathtype=folder >> a folder created
    path   = 'tests/_data/non_existing_folder_05'
    result = msk.data.make_path(path, pathtype='folder')
    assert result == True

    # ASSERTATION CHECK
    result = msk.data.ls('tests/_data', fmt='dict')
    assert len(result['folder']) == 4
    assert len(result['file'  ]) == 0
    assert ('non_existing_folder_04' in result['folder']) == False

    # reset content
    result = msk.data.rm('tests/_data')
    assert result == True
    result = msk.data.ls('tests/')
    assert ('_data' in result) == False


# -----------------------------
# SAVE AND LOAD CONTENT
# -----------------------------
def test_save_non_content_and_get_type_error():
    """ if content is None then create and empty file"""
    # resest the folder
    Mock.make_sure_the_folder_removed('tests/_data')
    # create a file
    path    = 'tests/_data/saved_content.txt'
    content = None
    msk.data.save(content, path)
    # check the file exist
    result  = msk.data.exist(path)
    assert result == True
    # content is empty string
    content = msk.data.load(path)
    assert content == ['']
    # resest the folder
    Mock.make_sure_the_folder_removed('tests/_data')


def test_load_incorrect_path():
    path   = 'tests/_data/content.txt'
    fd, fn = msk.data.path_split(path)
    Mock.make_sure_the_folder_removed(fd)

    with raises(FileNotFoundError):
        content = msk.data.load(path)


def test_save_and_load_object():
    # resest the folder
    content_folder = 'tests/_data'
    Mock.make_sure_the_folder_removed(content_folder)
    # create a dict content
    path    = f'{content_folder}/dict.bin'
    content = {'type': 'object'}
    msk.data.save_object(content, path)
    # check the file exist
    result  = msk.data.exist(path)
    assert result == True
    # content is empty string
    loaded_dict = msk.data.load_object(path)
    assert loaded_dict == {'type': 'object'}

    # create a list content
    path    = f'{content_folder}/list.bin'
    content = [1,2,3,4]
    msk.data.save_object(content, path)
    # check the file exist
    result  = msk.data.exist(path)
    assert result == True
    # content is empty string
    loaded_list = msk.data.load_object(path)
    assert loaded_list == [1,2,3,4]

    # other asserts
    ls_list = msk.data.ls(content_folder)
    assert len(ls_list) == 2
    assert ('dict.bin' in ls_list) == True
    assert ('list.bin' in ls_list) == True

    # resest the folder
    Mock.make_sure_the_folder_removed(content_folder)
    result  = msk.data.exist(content_folder)
    assert result == False


def test_save_and_load_plain_text():
    # resest the folder
    content_folder = 'tests/_data'
    Mock.make_sure_the_folder_removed(content_folder)

    # create a plain_text content
    path    = f'{content_folder}/one_line.txt'
    content = 'hello'
    msk.data.save(content, path)
    # check the file exist
    result  = msk.data.exist(path)
    assert result == True
    # content is 'hello'
    loaded_content = msk.data.load(path)
    assert loaded_content == ['hello']

    # create a list of string content
    path    = f'{content_folder}/multiple_lines.txt'
    content = ['hello', 'world']
    msk.data.save(content, path)
    # check the file exist
    result  = msk.data.exist(path)
    assert result == True
    # content is ['hello' 'world']
    loaded_content = msk.data.load(path)
    assert loaded_content == ['hello', 'world']

    # other asserts
    ls_list = msk.data.ls(content_folder)
    assert len(ls_list) == 2
    assert ('one_line.txt'       in ls_list) == True
    assert ('multiple_lines.txt' in ls_list) == True

    # resest the folder
    Mock.make_sure_the_folder_removed(content_folder)
    result  = msk.data.exist(content_folder)
    assert result == False


# -----------------------------
# LOAD CONTENT FROM FILE LIST
# -----------------------------
def test_load_list_of_files():
    # create list of files - 2 files
    folder_path = 'tests/_data'
    contents    = { 'f1.txt': ['a01', 'a02', ' ', 'a03', 'a4'],
                    'f2.txt': ['b01', 'b02', 'a03'] }
    Mock.make_content_files(folder_path, contents)

    # check if the files created properly
    f_list = [f'{folder_path}/{f}' for f in contents.keys()]

    # load conents from the list
    load_condition = lambda v: v.startswith('a')
    result1 = msk.data.load_file_list(f_list)           # ignore content that is whitespace
    result2 = msk.data.load_file_list(f_list, n_char=2) # ignore content that less than n_char
    result3 = msk.data.load_file_list(f_list, n_char=2, load_condition=load_condition) # overwrite n_char with load_condition

    # assertations
    for f in f_list: assert msk.data.exist(f) == True
    # - case result1
    assert len(result1) == 7
    assert ('a01' in result1) == True
    assert ('b02' in result1) == True
    assert ('b03' in result1) == False
    # - case result2
    assert len(result2) == 6
    assert ('a01' in result2) == True
    assert ('b02' in result2) == True
    assert ('a4'  in result2) == False
    # - case result3
    assert len(result3) == 5
    assert ('a01' in result3) == True
    assert ('b02' in result3) == False
    assert ('a4'  in result3) == True

    # reset the folder
    Mock.make_sure_the_folder_removed(folder_path)
    result = msk.data.exist(folder_path)
    assert result == False


# -----------------------------
# LOAD CATEGORY MAPPER CONTENT 
# -----------------------------
def test_category_mapper():
    # create list of files - 2 files
    folder_path     = 'tests/_data'
    contents = {'animal.txt': ['mammal,dog,cat', 'reptile,snake, croc, komodo'],
                'number.txt': ['digit,1,2,3','eng,one, two , three, twenty four '],
                'food.txt':   ['herb,', 'noodle, ramen, somen'] }
    Mock.make_content_files(folder_path, contents)

    # check if the files created properly
    f_list = [f'{folder_path}/{f}' for f in contents.keys()]

    # load contents from teh list
    cate, mapper  = msk.data.category_mapper(f_list)
    ls_list = msk.data.ls(folder_path)

    # assertations
    for f in f_list: assert msk.data.exist(f) == True
    assert len(ls_list) == 3
    assert set(ls_list) == set(list(contents.keys()))

    assert len(cate)       == 6
    assert cate['mammal' ] == ['mammal', 'dog', 'cat']
    assert cate['reptile'] == ['reptile', 'snake', 'croc', 'komodo']
    assert cate['herb'   ] == ['herb']
    assert ('two'         in cate['eng']) == True
    assert ('twenty four' in cate['eng']) == True

    assert len(mapper)    == 20
    assert mapper['two']  == 'eng'
    assert mapper['herb'] == 'herb'

    # reset the folder
    Mock.make_sure_the_folder_removed(folder_path)
    result = msk.data.exist(folder_path)
    assert result == False



import logging
import time
import os, pathlib, shutil
import dill
import yaml
import hashlib, base64, json, random

class timer:
    """ utility for calculate time duration
    """
    def __init__(self):
        """ init and tik() the timer """
        self._start = None
        self._stop  = None
        self.tik()

    def tik(self):
        self._start = time.time()
        return self._start

    def tok(self):
        self._stop  = time.time()
        return self._stop

    def taketime(seconds=None):
        """ static function for calc the time """
        if seconds:
            seconds  = seconds % (24 * 3600) 
            hour     = seconds // 3600
            seconds %= 3600
            minutes  = seconds // 60
            seconds %= 60
            return "%02d:%02d:%02d" % (hour, minutes, seconds) 
        else:
            return 'no given time (seconds)'


    def reset(self):
        """ reset start & stop timer to None, then tik() the timer """
        self._start = None
        self._stop  = None
        self.tik()

    def time(self, seconds=None):
        """ in-house function for calc timer """
        if seconds == None:
            seconds = self._stop - self._start if self._start and self._stop else None
 
        return timer.taketime(seconds)



class configure:
    def load(filename):
        """ load yaml file to configure """
        try:
            with open(filename, 'r') as f:
                conf = yaml.full_load(f)
            return conf

        except Exception as e:
            raise e


    def save(content, filename, sort_keys:bool=False):
        """ write data to yaml file configure """
        data.make_path(filename)
        with open(filename, 'w') as f:
            yaml.dump(content, f, allow_unicode=True, sort_keys=sort_keys)



class configure_loader:
    def __init__(self, f_configure):
        self.reset()
        self.load(f_configure)

    def reset(self):
        self.path    = None
        self.content = {}

    def load(self, f_configure:str=None):
        if f_configure:
            self.path    = f_configure
            self.content = configure.load(f_configure)

    def save(self):
        configure.save(self.content, self.path)

    def get(self, field):
        return self.content.get(field, None)



class data:
    def path_split(path:str) -> (str, str):
        """ split folder_path out of file_name
        """
        try:
            path   = path.strip()
            tokens = path.split('/')
            folder = '.' if len(tokens) == 1 else '/'.join(tokens[:-1])
            fname  = '' if tokens[-1] == '.' else tokens[-1]
            return folder, fname

        except Exception as e:
            raise e


    def path_type(path:str) -> str:
        """ check type [file, folder, None] of a given path
        """
        try:
            thing = pathlib.Path(path)
            if thing.exists():
                if   thing.is_dir():  return 'folder'
                elif thing.is_file(): return 'file'
                else:                return None
            else:
                return None

        except Exception as e:
            raise e


    def exist(path, pathtype:str='any') -> bool:
        """ check if given path exist """
        try:
            thing   = pathlib.Path(path)

            if   pathtype == 'any':    return thing.exists()
            elif pathtype == 'folder': return thing.is_dir()
            elif pathtype == 'file':   return thing.is_file()
            else: raise TypeError('pathtype should be ["any", "file" or "folder"]')

        except TypeError as e:
            raise e

        except Exception as e:
            raise e
            

    def make_path(path, pathtype:str='file') -> bool:
        """ make folder in the path, return True if success, else False 
            if the path is folder, adding / at the end would be preferable
        """
        # setup folder path
        if pathtype == 'file':
            _folder, _ = data.path_split(path)
        else:
            _folder    = path[:-1] if path.endswith('/') else path
        # create folder
        try: 
            #if len(folder) > 0: pathlib.Path(folder).mkdir(parents=True, exist_ok=True)
            if len(_folder) and _folder != '.': pathlib.Path(_folder).mkdir(parents=True, exist_ok=True)
        except Exception as e:
            print(f'error: {type(e)} - {str(e)}')
            return False
        return True


    def ls(path:str, fmt:str='list'):
        try:
            path   = '.' if path == '' else path
            result = {'folder': [], 'file': []}
            for obj in os.scandir(path):
                o_type = 'folder' if obj.is_dir() else 'file'
                result[o_type].append(obj.name)

            if fmt == 'list':
                out  = []
                out += [f'{item}/' for item in result['folder']]
                out += result['file']
                return out

            elif fmt == 'dict':
                return result

            else:
                raise Exception('something error')

        except Exception as e:
            raise e


    def rm(path:str, verbose:bool=False) -> bool:
        try:
            file_type = data.path_type(path)
            if   file_type == 'file':
                os.remove(path)
                if verbose: print(f'> file "{path}" removed')
                return True
            elif file_type == 'folder':
                shutil.rmtree(path)
                if verbose: print(f'> folder "{path}" removed')
                return True
            else:
                # file_type = None since the result of non-existing in path_type() 
                raise FileNotFoundError(f'"{path}" is not found')


        except FileNotFoundError as e:
            print(str(e))
            return False

        except Exception as e:
            print(f'> rm() error - {type(e)} - {str(e)}')
            return False


    def save_object(obj, filename, verbose:bool=True):
        """ save any data to dill object file """
        data.make_path(filename)
        with open(filename, 'wb') as f:
            dill.dump(obj, f)
            if verbose: print(f'> saving data to "{filename}"')


    def load_object(filename, verbose:bool=True):
        """ load content in a file to dill data object """
        if verbose: print(f'> loading data from "{filename}"')
        with open(filename, 'rb') as f:
            obj = dill.load(f)
        return obj


    def save(lines, filename, verbose:bool=True):
        """ save text in lines to a file """
        try:
            lines = '' if lines == None else lines
            lines = [lines] if type(lines) == str else lines
            data.make_path(filename)
            with open(filename, 'w') as f:
                for line in lines:
                    f.write(line+'\n')
                if verbose: print(f'> saving content to "{filename}"')

        except Exception as e:
            raise e


    def load(filename, loadtype='original', verbose:bool=True) -> list:
        """ load text content in a file """
        if verbose: print(f'> loading "{loadtype}" content from "{filename}"')
        with open(filename, 'r') as f:
            if loadtype == 'original': lines = [line[:-1]    for line in f.readlines()]
            else:                      lines = [line.strip() for line in f.readlines()]
        return lines


    def __assess_load_condition(load_condition=None, n_char:int=0):
        """ if load_condition is None, then the condition is True if its len() greater than n_char """
        if load_condition: return load_condition
        else:              return lambda v: True if len(v) > n_char else False

    def load_file_list(f_list, n_char:int=0, load_condition=None, verbose:bool=True) -> list:
        """ load content from a file list, each content line > n_char """
        # setup load_condition
        load_condition = data.__assess_load_condition(load_condition, n_char)
        # start the flow
        content = []
        n_fail  = 0
        for f in f_list:
            try:
                content += [v for v in data.load(f, loadtype='strip', verbose=verbose) if load_condition(v)]
            except Exception as e:
                n_fail  += 1
        if verbose: print(f'>>> load content from {len(f_list)} file(s) .. failed {n_fail} file(s)')
        return content


    def category_mapper(f_list, n_char:int=0, verbose:bool=True) -> (dict, dict):
        """ grouping a line of token list as a category
            return those categories and mapper
        """
        lines    = data.load_file_list(f_list, n_char, verbose=verbose)
        category = {line.split(',')[0].strip(): [token.strip() for token in line.split(',') if len(token.strip()) > 0] for line in lines}
        mapper   = {l: k for k, L in category.items() for l in L}
        return category, mapper
            
            

class hasher:
    def hash(obj, n:int=None, base:str='md5', force_hex:str=None) -> str:
        """ if obj type is dictionary, convert to string before hashing
            n is number of digit
            base: [md5, base64]
        """
        try:
            s_obj   = json.dumps(obj).encode() if type(obj) == dict else obj.encode()
            hashval = hashlib.md5(s_obj)
            if base == 'md5':
                hashval = hashval.hexdigest()
            else:
                hashval = base64.b64encode(hashval.digest())
                if force_hex: hashval = hashval.hex()
                hashval = hashval.decode('utf-8') if type(hashval)==bytes else hashval
            return hashval[:n]

        except AttributeError as e:
            print(f'> hasher.hash() error, obj type should be string or dict - {type(e)} - {str(e)}')
            raise e

        except TypeError as e:
            print(f'> hasher.hash() error, n type should be int or None - {type(e)} - {str(e)}')
            raise e

        except Exception as e:
            print(f'> hasher.hash() error - {type(e)} - {str(e)}')
            raise e


    def random_hash(prefix:str='', n:int=None, base:str='md5', force_hex:bool=None, seed:int=None) -> (str, str):
        """ randomly generate a key and its hash
            md5 - hex
            base64 - not hex
        """
        try:
            random.seed(seed)
            rd_seed = random.randint(0, 1_000_000_000_000)
            key  = f'{prefix}{rd_seed}'
            return key, hasher.hash(key, n, base=base, force_hex=force_hex)

        except TypeError as e:
            print(f'> hasher.random_hash() error, prefix type should be string, n type should be int or None - {type(e)} - {str(e)}')
            raise e

        except Exception as e:
            print(f'> hasher.random_hash() error - {type(e)} - {str(e)}')
            raise e

    def info():
        return 'random within range 0 - 1 billion | support base64 hex'



class param_tool:
    """ simple parser between dict and plain text
        this class does not smart enough to convert complex dict e.g., hierarchy dict
    """
    def validate_list(content:str, check_type:bool=True, plaintext:bool=False):
        """
        """
        if plaintext: return content
        chunks = content.split(',')
        result = []
        for chunk in chunks:
            try:
                chunk = chunk.strip()
                if check_type: chunk = eval(chunk)
            except NameError as e:
                ...
            except:
                ...

            finally:
                result.append(chunk)

        if len(chunks) == 1: return result[0]
        else:                return result


    def to_text(content:dict, delimeter='|', verbose:bool=False):
        """ input as dict e.g., content={ch:1, sch:3, subject: math}
            output will be 'ch=1|sch=3|subject=math'
        """
        if content is None:     return ''
        if len(content) == 0:   return ''
        parse_list = lambda L: ','.join([str(l) for l in L])
        return delimeter.join( [f'{k}={parse_list(v)}' if type(v) is list else f'{k}={v}' for k, v in content.items()] )


    def to_dict(content:str=None, delimeter='|', check_type:bool=True, plaintext:bool=False, verbose:bool=False):
        """ input as text e.g., content='ch=1|sch=3|subject=math'
            output will be dict {ch:1, sch:3, subject: math}
        """
        try:
            if content is None:             return {}
            if len(content.strip()) == 0:   return {}
            if verbose: print(f'> content: {content}')
            chunks = content.split(delimeter)
            params = {}
            for chunk in chunks:
                k, v      = chunk.split('=')
                k         = k.replace(' ', '')
                v         = v.strip()
                params[k] = param_tool.validate_list(v, check_type=check_type, plaintext=plaintext)
            return params

        except ValueError as e:
            print(f'> parse_tool.to_dict("{content}") with delimeter="{delimeter}" error - {type(e)} - {str(e)}')
            return {}

    def dict_to_key_val_list(content:dict, key='key', val='val'):
        try:
            return [{key: k, val: v} for k, v in content.items()]
        except Exception as e:
            print(f'> parse_tool.dict_to_key_val_list("{content}") with with key: {key}, val: {val}  error - {type(e)} - {str(e)}')
            raise e


    def key_val_list_to_dict(content:list, key='key', val='val'):
        try:
            return {it[key]: it[val] for it in content}
        except Exception as e:
            print(f'> parse_tool.key_val_list_to_dict("{content}") with with key: {key}, val: {val}  error - {type(e)} - {str(e)}')
            raise e





    def repack(content:str, delimeter:str='|', check_type:bool=True, plaintext:bool=False, verbose:bool=False, **kwargs):
        """ modify values in one-ine content
        """
        if not kwargs: return content
        # parse content to dict
        params = param_tool.to_dict(content, delimeter=delimeter, check_type=check_type, plaintext=plaintext,  verbose=verbose)
        # update params
        params.update(kwargs)
        # pack dict to one-line content
        return param_tool.to_text(params, delimeter=delimeter, verbose=verbose)





class log:
    import logging.handlers

    """ static class for creating logger instance """
    DEBUG    = 10
    INFO     = 20
    WARNING  = 30
    ERROR    = 40
    CRITICAL = 50

    FORMATTER= {k: k for k in ['nothing', 'minimal', 'basic', 'full']}

    def info(logger, message=''):
        """ print information of the given handler """
        print(f'--- {message} logger "{logger.name}" info ---')
        print(f'logger: {logger}')
        print(f'handlers: {len(logger.handlers)}')
        for hand in logger.handlers:
            print(f'  {hand}')
        print()

    def _validate_formatter(formatter:str='minimal'):
        """ formatter is a list of [nothing, minimal, basic, full] otherwise, nothing
        """
        if not formatter: formatter = 'minimal'
        formatter = formatter.lower()
        return log.FORMATTER.get(formatter, 'nothing')

    def _validate_when(when:str):
        """ modify the readable {when} to a format that logging can process """
        if when is None: return None
        # daily  -> midnight
        # weekly -> W0
        return 'midnight' if when == 'daily' else 'W0' if when == 'weekly' else when


    def GetHandler(filename=None, when=None, level:int=logging.DEBUG, formatter:str='minimal'):
        """ formatter will be in f-string format
            formatter: [nothing, minimal, basic, full, custom]
            o nothing: only message
            o minimal: time, message
            o basic:   level, name, message (exclude time)
            o full:    level, time, name, message
            o custom:  any format, but in f-string format only

            when: preset is [daily, weekly] or logging format e.g., W0
        """
        # validate format of the given formatter
        formatter = log._validate_formatter(formatter)
        if   formatter == 'nothing':      formatter = '{message}'
        elif formatter == 'minimal':      formatter = '{asctime} || {message}'
        elif formatter == 'basic':        formatter = '{asctime} || {name} || {message}'
        elif formatter == 'full':         formatter = '{levelname:<8} || {asctime} || {name} || {message}'

        fmt = logging.Formatter(formatter, style='{')
        if filename:
            # make sure the folder is exist
            data.make_path(filename)
            # create log_handler for file-logging
            w = log._validate_when(when)
            if w is not None:
                handler = logging.handlers.TimedRotatingFileHandler(filename, when=w, encoding='utf8')
            else:
                handler = logging.FileHandler(filename, encoding='utf8')
            
        else:
            handler = logging.StreamHandler()
        handler.setLevel(level)
        handler.setFormatter(fmt)
        return handler


    def clear_by_name(name:str):
        """ force clear the logger if already exist """
        if logging.getLogger(name).hasHandlers(): logging.getLogger(name).handlers.clear()


    def GetLogger(name='muuusiiik', filename=None, when=None, level=logging.DEBUG, formatter='minimal'):
        """ generate Logger """
        # manage logger
        if not name: name = __name__
        # clear the existing logger if already exist
        log.clear_by_name(name)
        # create root logger
        logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        # manage handler
        handler = log.GetHandler(filename=filename, when=when, level=level, formatter=formatter) 
        logger.addHandler(handler)
        return logger



def json_update(data, input_value, output_value, parent_node='', parents=['']):
    """
    input parameters
    data:dict = dict type variable.
    input_value:dict = dict key and value that you want to check matching.
    output:dict = dict key and value that you want to upate value.
    parent_node:str = parents key, leave it blank if you need to update only 1st level of dict.
    
    example:
    ----------------------
    data = {'botname': 'apollo',
     'linetoken': 'nXexNPQNE/7GU+TuUITd3eOI3xFKupMgs4l2GBr3s9WpOPFCtLGMa1',
     'webhook': 'https://abdul.in.th/lite/wh/komyo/1dafdacbdc9a8fd434928ad2bb8d0b14.php',
     'apikey': 'xxx',
     'botid': '1dafdacbdc9a8fd434928ad2bb8d0b14',
     'ownerid': 'komyo',
     'category': '',
     'objective': '',
     'knowledge': [{'name': 'Conversation',
       'id': '1Crt9HIHzXG1Wv_wDD48QGoj400NLUBuU0YQiIAG2OXs',
       'type': 'cu'},
      {'name': 'FAQ',
       'id': '1Crt9HIHzXG1Wv_wDD48QGoj400NLUBuU0YQiIAG2OXs',
       'type': 'faq'}],
     'idea': {'ownerid': 'komyo',
      'type': 'cu',
      'more_idea': {'type': 'cu', 'types': ['sam', 'cu', 2]}}}
      
    input_value = {'ownerid': 'komyo'}
    
    output_value = {'type': 'sam'}
    parent_node = 'idea' 
    
    
    output :
    ----------------------
    {'botname': 'apollo',
     'linetoken': 'nXexNPQNE/7GU+TuUITd3eOI3xFKupMgs4l2GBr3s9WpOPFCtLGMa1',
     'webhook': 'https://abdul.in.th/lite/wh/komyo/1dafdacbdc9a8fd434928ad2bb8d0b14.php',
     'apikey': 'xxx',
     'botid': '1dafdacbdc9a8fd434928ad2bb8d0b14',
     'ownerid': 'komyo',
     'category': '',
     'objective': '',
     'knowledge': [{'name': 'Conversation',
       'id': '1Crt9HIHzXG1Wv_wDD48QGoj400NLUBuU0YQiIAG2OXs',
       'type': 'cu'},
      {'name': 'FAQ',
       'id': '1Crt9HIHzXG1Wv_wDD48QGoj400NLUBuU0YQiIAG2OXs',
       'type': 'faq'}],
     'idea': [{'ownerid': 'komyo',
       'type': 'sam',
       'more_idea': {'type': 'cu', 'types': ['sam', 'cu', 2]}}]}
    """
    output_key = list(output_value.keys())[0]
    input_key = list(input_value.keys())[0]
    
    for key, value in data.items():
        # check key is equal to input key 
        if key == input_key and value == input_value[input_key] and \
                output_key in data.keys() and parent_node in parents:
            #
            data[output_key] = output_value[output_key]

        elif isinstance(value, list):
            for idx, item in enumerate(value):
                # check type dict
                if isinstance(item, dict):
                    json_update(item, input_value, output_value, parent_node, list(data.keys()))
                    
    return data



from setuptools import setup
from setuptools import find_packages

README = ''.join([line for line in open("README.md").readlines()])
setup(
    name="muuusiiik",
    version="0.0.34",
    author="@muuusiiik",
    author_email="muuusiiikd@gmail.com",
    description="simple utilities",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/muuusiiik/utility",
    python_requires=">=3.7",
    license="MIT",
     classifiers=[
         "License :: OSI Approved :: MIT License",
         "Programming Language :: Python :: 3",
         "Programming Language :: Python :: 3.7",
     ],
     #packages=["muuusiiik"],
     packages=find_packages("."),
     namespace_package=['muuusiiik'],

     # DEAL WITH DATA
     #package_data={"vocab": ["vocab/*"]}
     #include_package_data=True,

     # DEPENDENCIES PACKAGES
     install_requires=["dill", "PyYaml"],
 )




import muuusiiik.util as msk

#===============
# VALIDATE LIST
#===============
def test_param_tool_validate_list_normal_string():
    content  = 'hello'
    result   = msk.param_tool.validate_list(content)
    assert result == 'hello'

def test_param_tool_validate_list_spacing_string():
    content  = ' hello '
    result   = msk.param_tool.validate_list(content)
    assert result == 'hello'

def test_param_tool_validate_list_string_in_list():
    content  = 'hello, world '
    result   = msk.param_tool.validate_list(content)
    assert result == ['hello', 'world']

def test_param_tool_validate_list_number_in_list():
    content  = ' 1.0, 2, 3.4'
    result   = msk.param_tool.validate_list(content)
    assert result == [1.0, 2, 3.4]

def test_param_tool_validate_list_mixed_number_in_list():
    """ number will be converted as number
    """
    content  = ' 1.0, two, 3.4'
    result   = msk.param_tool.validate_list(content)
    assert result == [1.0, 'two', 3.4]

def test_param_tool_validate_list_mixed_number_in_list_without_check_type():
    """ number will be converted as string
    """
    content  = ' 1.0, two, 3.4'
    result   = msk.param_tool.validate_list(content, check_type=False)
    assert result == ['1.0', 'two', '3.4']

def test_param_tool_validate_list_plain_text():
    """ keep it is plain-text (do nothing)
    """
    content  = 'qno:+2,sc=3|qno:3'
    result   = msk.param_tool.validate_list(content, check_type=False, plaintext=True)
    assert result == content

#=================
# CONVERT CONTENT
#=================
def test_param_tool_convert_dict_to_text():
    content  = {'ch': 1, 'sch': 1, 'done': [1,2,3,4], 'session':'init'}
    result   = msk.param_tool.to_text(content, delimeter='|')
    expected = 'ch=1|sch=1|done=1,2,3,4|session=init'
    assert result == expected


def test_param_tool_convert_text_to_dict():
    content  = 'ch=1|sch=1|done=1,2,3,4|session=init'
    result   = msk.param_tool.to_dict(content, delimeter='|')
    expected = {'ch': 1, 'sch': 1, 'done': [1,2,3,4], 'session':'init'}
    assert result == expected


#================
# REPACK CONTENT
#================
def test_param_tool_repack_param_to_content():
    content  = 'ch=1|sch=1|session=init'
    kwargs   = {'done': [1,2,3,4], 'session': 'next step'}
    result   = msk.param_tool.repack(content, **kwargs)
    expected = 'ch=1|sch=1|session=next step|done=1,2,3,4'
    assert result == expected




import muuusiiik.util as msk
import time

def test_timer_run_when_init():
    """ round time() to int for approximation """
    now = int( time.time() )
    t   = msk.timer()
    tik = int( t._start )
    assert tik == now


def test_timer_run_when_reset():
    """ round time() to int for approximation """
    now = int( time.time() )
    t   = msk.timer()
    t.reset()
    tik = int( t._start )
    assert tik == now


def test_time_function_by_given_duration_in_second():
    t      = msk.timer()
    result = t.time(305)
    assert result == '00:05:05'


def test_taketime_function_for_being_called_by_the_class():
    result = msk.timer.taketime(305)
    assert result == '00:05:05'






import muuusiiik.util as msk
import random
from   pytest import raises


class RandomMock:
    _tb = {1: 1, 2: 2, 3: 3}
    def n():
        rd = random.choice([1,2,3])
        return RandomMock._tb[rd]


# -----------------------------
# TEST CASES FOR HASHER.HASH()
# -----------------------------

def test_hash_string():
    obj    = 'demo text'
    result = msk.hasher.hash(obj)
    assert type(obj) == str
    assert result    == '5c40f9622c9489a25dd71a15374a04fe'


def test_hash_dict():
    obj    = {'type': 'dict', 'value': 'some value'}
    result = msk.hasher.hash(obj)
    assert type(obj) == dict
    assert result    == '2f7a3110878f1c45dcafa9f62b428171'


def test_hash_dict_with_array():
    obj    = {'type': 'dict', 'value': [1, 2, 3, '4']}
    result = msk.hasher.hash(obj)
    assert type(obj) == dict
    assert result    == '84c9de9707639d7a2d28140add70dc97'


def test_hash_dict_with_spaceing_array_should_have_the_same_result_as_normal_array():
    obj    = {'type': 'dict', 'value': [1,   2,     3,     '4']}
    result = msk.hasher.hash(obj)
    assert type(obj) == dict
    assert result    == '84c9de9707639d7a2d28140add70dc97'


def test_hash_for_4_digit():
    obj    = 'demo text'
    result = msk.hasher.hash(obj, n=4)
    assert type(obj)   == str
    assert len(result) == 4
    assert result      == '5c40'


def test_hash_for_n_digits():
    n_digit = RandomMock.n()
    obj     = 'demo text'
    result  = msk.hasher.hash(obj, n=n_digit)
    assert type(obj)   == str
    assert len(result) == n_digit
    assert result      == '5c40f9622c9489a25dd71a15374a04fe'[:n_digit]


def test_exception_attribute_error_when_obj_is_int():
    obj     = 404
    with raises(AttributeError):
        result  = msk.hasher.hash(obj)


def test_exception_attribute_error_when_obj_is_list():
    obj     = [404]
    with raises(AttributeError):
        result  = msk.hasher.hash(obj)


def test_exception_type_error_when_n_is_not_int():
    obj     = 'demo text'
    with raises(TypeError):
        result  = msk.hasher.hash(obj, n=[4])


# ------------------------------------
# TEST CASES FOR HASHER.RANDOM_HASH()
# ------------------------------------

def test_random_hash_return_tuple_of_string():
    prefix = 'this_is_prefix'
    result = msk.hasher.random_hash(prefix)
    assert len(result) == 2


def test_random_hash_return_4_digits():
    prefix = 'this_is_prefix'
    n      = 4
    result = msk.hasher.random_hash(prefix, n=n)
    assert len(result[1]) == n


def test_random_hash_return_prefix_concated_with_two_digit():
    prefix = 'this_is_prefix'
    hash_key, hash_value = msk.hasher.random_hash(prefix)
    assert hash_key[:-2]           == prefix
    assert hash_key[-2:].isdigit() == True


def test_random_hash_also_raise_type_error_if_prefix_is_not_string():
    with raises(TypeError):
        prefix = 404
        msk.hasher.hash(prefix=prefix)


def test_random_hash_also_raise_type_error_if_n_is_not_int():
    with raises(TypeError):
        n = '4'
        msk.hasher.random_hash(n=n)
